# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Optimizers for use with JAX.

This short module contains some convenient optimizer definitions, specifically
initialization and update functions, which can be used with ndarrays or
arbitrarily-nested tuple/list/dicts of ndarrays.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import operator
import functools

import jax.numpy as np
from jax.core import pack
from jax.tree_util import tree_map, tree_multimap


def optimizer(opt_maker):
  """Decorator to make an optimizer map over tuple/list/dict containers."""
  @functools.wraps(opt_maker)
  def tree_opt_maker(*args, **kwargs):
    init_fun, update_fun = opt_maker(*args, **kwargs)

    @functools.wraps(init_fun)
    def fmapped_init_fun(x0_tree):
      return tree_map(lambda x0: pack(init_fun(x0)), x0_tree)

    @functools.wraps(update_fun)
    def fmapped_update_fun(i, grad_tree, state_tree):
      update = lambda g, state: pack(update_fun(i, g, *state))
      return tree_multimap(update, grad_tree, state_tree)

    return fmapped_init_fun, fmapped_update_fun
  return tree_opt_maker

def iterate(state_tree):
  """Extract the current iterate from an optimizer state."""
  return tree_map(lambda state: tuple(state)[0], state_tree)
get_params = iterate

# optimizers

@optimizer
def sgd(step_size):
  """Construct init and update step functions for stochastic gradient descent.

  Args:
    step_size: positive scalar, or a callable representing a step size schedule
      that maps the iteration index to positive scalar.

  Returns:
    An (init_fun, update_fun) pair.
  """
  step_size = make_schedule(step_size)
  def init_fun(x0):
    return (x0,)
  def update_fun(i, g, x):
    return (x - step_size(i) * g,)
  return init_fun, update_fun

@optimizer
def momentum(step_size, mass):
  """Construct init and update step functions for SGD with Nesterov momentum.

  Args:
    step_size: positive scalar, or a callable representing a step size schedule
      that maps the iteration index to positive scalar.

  Returns:
    An (init_fun, update_fun) pair.
  """
  step_size = make_schedule(step_size)
  def init_fun(x0):
    v0 = np.zeros_like(x0)
    return x0, v0
  def update_fun(i, g, x, velocity):
    velocity = mass * velocity - (1. - mass) * g
    x = x + step_size(i) * velocity
    return x, velocity
  return init_fun, update_fun

@optimizer
def rmsprop(step_size, gamma=0.9, eps=1e-8):
  """Construct init and update step functions for RMSProp.

  Args:
    step_size: positive scalar, or a callable representing a step size schedule
      that maps the iteration index to positive scalar.

  Returns:
    An (init_fun, update_fun) pair.
  """
  step_size = make_schedule(step_size)
  def init_fun(x0):
    avg_sq_grad = np.ones_like(x0)
    return x0, avg_sq_grad
  def update_fun(i, g, x, avg_sq_grad):
    avg_sq_grad = avg_sq_grad * gamma + g**2 * (1. - gamma)
    x = x - step_size(i) * g / (np.sqrt(avg_sq_grad) + eps)
    return x, avg_sq_grad
  return init_fun, update_fun

@optimizer
def adam(step_size, b1=0.9, b2=0.999, eps=1e-8):
  """Construct init and update step functions for Adam.

  Args:
    step_size: positive scalar, or a callable representing a step size schedule
      that maps the iteration index to positive scalar.
    b1: optional, a positive scalar value for beta_1, the exponential decay rate
      for the first moment estimates (default 0.9).
    b2: optional, a positive scalar value for beta_2, the exponential decay rate
      for the second moment estimates (default 0.999).
    eps: optional, a positive scalar value for epsilon, a small constant for
      numerical stability (default 1e-8).

  Returns:
    An (init_fun, update_fun) pair.
  """
  step_size = make_schedule(step_size)
  def init_fun(x0):
    m0 = np.zeros_like(x0)
    v0 = np.zeros_like(x0)
    return x0, m0, v0
  def update_fun(i, g, x, m, v):
    m = (1 - b1) * g + b1 * m  # First  moment estimate.
    v = (1 - b2) * (g ** 2) + b2 * v  # Second moment estimate.
    mhat = m / (1 - b1 ** (i + 1))  # Bias correction.
    vhat = v / (1 - b2 ** (i + 1))
    x = x - step_size(i) * mhat / (np.sqrt(vhat) + eps)
    return x, m, v
  return init_fun, update_fun

# learning rate schedules

def constant(step_size):
  def schedule(i):
    return step_size
  return schedule

def exponential_decay(step_size, decay_steps, decay_rate):
  def schedule(i):
    return step_size * decay_rate ** (i / decay_steps)
  return schedule

def inverse_time_decay(step_size, decay_steps, decay_rate, staircase=False):
  if staircase:
    def schedule(i):
      return step_size / (1 + decay_rate * np.floor(i / decay_steps))
  else:
    def schedule(i):
      return step_size / (1 + decay_rate * i / decay_steps)
  return schedule

def piecewise_constant(boundaries, values):
  boundaries = np.array(boundaries)
  values = np.array(values)
  if not boundaries.ndim == values.ndim == 1:
    raise ValueError("boundaries and values must be sequences")
  if not boundaries.shape[0] == values.shape[0] - 1:
    raise ValueError("boundaries length must be one longer than values length")

  def schedule(i):
    return values[np.sum(i > boundaries)]
  return schedule

def make_schedule(scalar_or_schedule_fun):
  if callable(scalar_or_schedule_fun):
    return scalar_or_schedule_fun
  elif np.ndim(scalar_or_schedule_fun) == 0:
    return constant(scalar_or_schedule_fun)
  else:
    raise TypeError(type(scalar_or_schedule_fun))



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import namedtuple
import itertools as it
import numpy as onp
import operator as op
import six
from six.moves import xrange

from ..config import flags
from .. import core
from .. import ad_util
from ..abstract_arrays import ConcreteArray, ShapedArray, make_shaped_array, array_types
from ..core import AbstractTuple, JaxTuple, pack, valid_jaxtype
from ..util import partial, partialmethod, memoize, unzip2, concatenate, safe_map, prod
from ..linear_util import transformation_with_aux, memoize as linear_memoize
from ..lib import xla_bridge as xb
from .partial_eval import trace_to_subjaxpr, merge_pvals, JaxprTrace, PartialVal

FLAGS = flags.FLAGS
flags.DEFINE_bool('jax_device_values', True, 'Enable device-persistent values.')

map = safe_map

def apply_primitive(prim, *args, **kwargs):
  abstract_args = map(abstractify, args)
  compiled_fun = xla_primitive_callable(prim, *abstract_args, **kwargs)
  return compiled_fun(*args)

@memoize
def xla_primitive_callable(prim, *abstract_args, **kwargs):
  shapes = map(xla_shape, abstract_args)
  built_c = primitive_computation(prim, *shapes, **kwargs)
  result_shape = xla_shape_to_result_shape(built_c.GetReturnValueShape())
  handle_result = result_handler(result_shape)
  compiled = built_c.Compile(shapes, xb.get_compile_options())
  return partial(execute_compiled_primitive, compiled, handle_result)

@memoize
def primitive_computation(prim, *shapes, **kwargs):
  c = xb.make_computation_builder("primitive_computation")
  xla_args = map(c.ParameterWithShape, shapes)
  xla_result = translation_rule(prim)(c, *xla_args, **kwargs)
  try:
    return c.Build()
  except RuntimeError as e:
    # try for a better error message by using the abstract_eval checks
    prim.abstract_eval(*map(aval_from_xla_shape, shapes), **kwargs)
    raise e

def aval_from_xla_shape(shape):
  return ShapedArray(shape.dimensions(), shape.element_type())

def execute_compiled_primitive(compiled, result_handler, *args):
  input_bufs = [device_put(canonicalize_pyval_dtype(x)) for x in args]
  return result_handler(compiled.Execute(input_bufs, not core.skip_checks))

def device_put(x):
  if type(x) is DeviceArray:
    return x.device_buffer
  else:
    return xb.device_put(x)  # can round-trip elements of tuples here

def result_handler(result_shape):
  if type(result_shape) is ResultArray and FLAGS.jax_device_values:
    def handle_result(device_buffer):
      return DeviceArray(device_buffer, *result_shape)
  elif type(result_shape) is ResultArray:
    def handle_result(device_buffer):
      return onp.asarray(DeviceArray(device_buffer, *result_shape))
  elif type(result_shape) is ResultTuple:
    handlers = list(map(result_handler, result_shape))
    def handle_result(device_buffer):
      bufs = device_buffer.destructure()
      return JaxTuple(handler(buf) for handler, buf in zip(handlers, bufs))
  else:
    raise TypeError(type(result_shape))
  return handle_result

def xla_shape_to_result_shape(xla_shape):
  if xla_shape.is_tuple():
    return ResultTuple(map(xla_shape_to_result_shape, xla_shape.tuple_shapes()))
  else:
    shape, dtype = xla_shape.dimensions(), xla_shape.element_type()
    ndim, size = len(shape), prod(shape)
    return ResultArray((shape, dtype, ndim, size))
class ResultTuple(tuple): pass
class ResultArray(tuple): pass


def compile_jaxpr(jaxpr, const_vals, *abstract_args):
  arg_shapes = list(map(xla_shape, abstract_args))
  built_c = jaxpr_computation(jaxpr, const_vals, (), *arg_shapes)
  result_shape = xla_shape_to_result_shape(built_c.GetReturnValueShape())
  return built_c.Compile(arg_shapes, xb.get_compile_options()), result_shape

def jaxpr_computation(jaxpr, const_vals, freevar_shapes, *arg_shapes):
  c = xb.make_computation_builder("jaxpr_computation")

  def read(v):
    return env[v]

  def write(v, node):
    assert node is not None
    env[v] = node

  env = {}
  consts_env = dict(zip(jaxpr.constvars, const_vals))
  write(core.unitvar, c.Tuple())
  map(write, jaxpr.constvars, map(c.Constant, const_vals))
  map(write, jaxpr.freevars, map(c.ParameterWithShape, freevar_shapes))
  map(write, jaxpr.invars, map(c.ParameterWithShape, arg_shapes))
  for eqn in jaxpr.eqns:
    in_nodes = map(read, eqn.invars)
    in_shapes = map(c.GetShape, in_nodes)
    subcs = [jaxpr_computation(subjaxpr,
                               [consts_env[b] for b in const_bindings],
                               map(c.GetShape, map(read, freevar_bindings)),
                               *in_shapes)
             for subjaxpr, const_bindings, freevar_bindings
             in eqn.bound_subjaxprs]
    subfuns = [(subc, tuple(map(read, freevar_bindings)))
               for subc, (_, _, freevar_bindings)
               in zip(subcs, eqn.bound_subjaxprs)]
    ans = translation_rule(eqn.primitive)(c, *(subfuns + in_nodes), **eqn.params)
    out_nodes = xla_destructure(c, ans) if eqn.destructure else [ans]
    map(write, eqn.outvars, out_nodes)
  return c.Build(read(jaxpr.outvar))

def xla_destructure(c, ans):
  num_elements = len(c.GetShape(ans).tuple_shapes())
  return [c.GetTupleElement(ans, i) for i in range(num_elements)]

def unit_constant(c, val):
  assert not val  # must be unit
  return c.Tuple()
xb.register_constant_handler(JaxTuple, unit_constant)

def translation_rule(p):
  try:
    return translations[p]
  except KeyError:
    raise NotImplementedError(
        "XLA translation rule for '{}' not implemented".format(p))


translations = {}

translations[core.pack_p] = lambda c, *xs: c.Tuple(*xs)
translations[core.call_p] = lambda c, subc_a1, *a2: c.Call(subc_a1[0],
                                                           subc_a1[1] + a2)
translations[core.identity_p] = lambda c, x: x

# TODO(mattjj): zeros_like and add_jaxvals should handle any jaxval
def zeros_like_translation_rule(c, x):
  x_shape = c.GetShape(x)
  return c.Broadcast(c.Constant(onp.array(0, x_shape.element_type())),
                     x_shape.dimensions())
translations[ad_util.zeros_like_p] = zeros_like_translation_rule
translations[ad_util.add_jaxvals_p] = lambda c, x, y: c.Add(x, y)


def canonicalize_pyval_dtype(x):
  try:
    return canonicalize_dtype_handlers[type(x)](x)
  except KeyError:
    msg = "No canonicalize handler registered for type: {}"
    raise TypeError(msg.format(type(x)))

canonicalize_dtype_handlers = {}

def canonicalize_tuple_dtype(tup):
  return JaxTuple(map(canonicalize_pyval_dtype, tup))
canonicalize_dtype_handlers[JaxTuple] = canonicalize_tuple_dtype

def canonicalize_ndarray_dtype(x):
  return onp.asarray(x, xb.canonicalize_dtype(onp.result_type(x)))

for t in array_types:
  canonicalize_dtype_handlers[t] = canonicalize_ndarray_dtype

def identity(x): return x


def abstractify(x):
  try:
    return pytype_aval_mappings[type(x)](x)
  except KeyError:
    raise TypeError("No abstraction handler for type: {}".format(type(x)))

pytype_aval_mappings = {}

def abstractify_tuple(tup):
  return AbstractTuple(tuple(map(abstractify, tup)))
pytype_aval_mappings[JaxTuple] = abstractify_tuple

for t in array_types:
  pytype_aval_mappings[t] = make_shaped_array


class DeviceValue(object):
  __slots__ = ["device_buffer"]
  def __init__(self, device_buffer):
    self.device_buffer = device_buffer

def forward_method(attrname, self, fun, *args):
  return fun(getattr(self, attrname), *args)
forward_to_value = partial(forward_method, "_value")

class DeviceArray(DeviceValue):
  __slots__ = ["shape", "dtype", "ndim", "size", "_npy_value"]
  __array_priority__ = 100.

  def __init__(self, device_buffer, shape, dtype, ndim, size):
    self.device_buffer = device_buffer
    self.shape = shape
    self.dtype = dtype
    self.ndim = ndim
    self.size = size
    self._npy_value = None

  @property
  def _value(self):
    if self._npy_value is None:
      self._npy_value = self.device_buffer.to_py()
      try:
        self._npy_value.flags.writeable = False
      except AttributeError:
        # TODO(mattjj): bug with C64 on TPU backend, C64 values returned as pair
        if onp.issubdtype(self.dtype, onp.complexfloating):
          a, b = self._npy_value
          npy_value = onp.stack([a, b], -1).view(self.dtype).reshape(self.shape)
          npy_value.flags.writeable = False
          self._npy_value = npy_value
        else:
          raise
    return self._npy_value

  def copy(self):
    """Returns an ndarray (backed by host memory, not device memory)."""
    return onp.asarray(self)

  def __repr__(self):
    shape_str = ",".join(map(str, self.shape))
    return "DeviceArray{{{}[{}]}}".format(self.dtype.name, shape_str)

  def __len__(self):
    try:
      return self.shape[0]
    except IndexError:
      raise TypeError("len() of unsized object")  # same as numpy error

  def __iter__(self):
    if self.ndim == 0:
      raise TypeError("iteration over a 0-d array")  # same as numpy error
    else:
      return (self[i] for i in xrange(self.shape[0]))

  def __reversed__(self):
    if self.ndim == 0:
      raise TypeError("iteration over a 0-d array")
    else:
      return (self[i] for i in xrange(self.shape[0] - 1, -1, -1))

  def __format__(self, format_spec):
    # Simulates behavior of https://github.com/numpy/numpy/pull/9883
    if self.ndim == 0:
      return format(self._value[()], format_spec)
    else:
      return format(self._value, format_spec)

  __array__ = partialmethod(forward_to_value, onp.asarray)
  __str__ = partialmethod(forward_to_value, str)
  __bool__ = __nonzero__ = partialmethod(forward_to_value, bool)
  __float__ = partialmethod(forward_to_value, float)
  __int__ = partialmethod(forward_to_value, int)
  if six.PY2:
    __long__ = partialmethod(forward_to_value, long)
  __complex__ = partialmethod(forward_to_value, complex)
  __hex__ = partialmethod(forward_to_value, hex)
  __oct__ = partialmethod(forward_to_value, oct)

  # pickle saves and loads just like an ndarray
  __reduce__ = partialmethod(forward_to_value, op.methodcaller("__reduce__"))

  # clobbered when jax.numpy is imported, but useful in tests
  def __eq__(self, other): return self._value == other

  def __hash__(self):
    # TODO(mattjj): this is not semantically correct because it is possible
    # __eq__ is true for values with unequal __hash__ values. However, the
    # main use case at the moment is memoization for which false negatives are
    # fine.
    return id(self)


core.pytype_aval_mappings[DeviceArray] = ConcreteArray
pytype_aval_mappings[DeviceArray] = make_shaped_array
canonicalize_dtype_handlers[DeviceArray] = identity
xb.register_constant_handler(DeviceArray,
                             lambda c, val: c.Constant(onp.asarray(val)))


def xla_shape(x):
  try:
    return xb.Shape.array_shape(x.dtype, x.shape)
  except AttributeError:
    if type(x) in (core.AbstractTuple, core.JaxTuple):
      return xb.Shape.tuple_shape(tuple(map(xla_shape, x)))
    else:
      raise TypeError(type(x))


# For callable XLA Computations (as opposed to, e.g., Computations used in the
# body of a While) we flatten functions to take multiple array arguments (no
# tuple arguments) and return either an array output or a flat tuple output that
# is immediately destructured. This flattening avoids the need for the runtime
# to manage multiple references to DeviceValues caused by tuple membership
# (since the XLA runtime depends on single-ownership, rather than e.g.
# refcounting). In particular, we don't have a DeviceTuple representation, and
# instead, for values returned to the user, always destructure tuples.
# The code here is similar to that in tree_util, but is meant to flatten
# JaxTuple trees only.

@transformation_with_aux
def flatten_fun(in_trees, *flat_args):
  jtuple_trees = tuple(map(partial(build_tree, iter(flat_args)), in_trees))
  ans = yield jtuple_trees
  if type(ans) is JaxTuple:
    ans_flat, out_tree = tree_flatten(ans)
    yield pack(ans_flat), out_tree
  else:
    yield ans, leaf

def tree_flatten(maybe_tree):
  if type(maybe_tree) is JaxTuple:
    flat_children, child_specs = unzip2(map(tree_flatten, maybe_tree))
    return it.chain.from_iterable(flat_children), JTupleTreeDef(child_specs)
  elif core.skip_checks or valid_jaxtype(maybe_tree):
    return [maybe_tree], leaf
  else:
    raise TypeError(type(maybe_tree))

JTupleTreeDef = namedtuple("JTupleTreeDef", ["child_specs"])

class Leaf(object):
  def __repr__(self):
    return '*'
leaf = Leaf()

def build_tree(xs, tree_spec):
  if tree_spec is leaf:
    return next(xs)
  elif type(tree_spec) is JTupleTreeDef:
    return pack(map(partial(build_tree, xs), tree_spec.child_specs))
  else:
    raise TypeError(type(tree_spec))


def xla_call_impl(fun, *args):
  flat_args, in_trees = unzip2(map(tree_flatten, args))
  flat_args = concatenate(flat_args)
  fun, out_tree = flatten_fun(fun, in_trees)

  compiled_fun = xla_callable(fun, *map(abstractify, flat_args))
  flat_ans = compiled_fun(*flat_args)

  if out_tree() is leaf:
    return flat_ans
  else:
    return build_tree(iter(flat_ans), out_tree())

@linear_memoize
def xla_callable(fun, *abstract_args):
  with core.new_master(JaxprTrace, True) as master:
    pvals = [PartialVal((aval, core.unit)) for aval in abstract_args]
    jaxpr, (pval, consts, env) = trace_to_subjaxpr(fun, master).call_wrapped(pvals)
    assert not env  # no subtraces here (though cond might eventually need them)
    compiled, result_shape = compile_jaxpr(jaxpr, consts, *abstract_args)
    del master, pvals, consts, jaxpr, env
    handle_result = result_handler(result_shape)
    return partial(execute_compiled, compiled, pval, handle_result)

def execute_compiled(compiled, pval, handle_result, *args):
  input_bufs = [device_put(canonicalize_pyval_dtype(x)) for x in args]
  out_buf = compiled.Execute(input_bufs, not core.skip_checks)
  return merge_pvals(handle_result(out_buf), pval)


xla_call_p = core.Primitive('xla_call')
xla_call = partial(core.call_bind, xla_call_p)
xla_call_p.def_custom_bind(xla_call)
xla_call_p.def_impl(xla_call_impl)

translations[xla_call_p] = lambda c, subc_a1, *a2: c.Call(subc_a1[0],
                                                          subc_a1[1] + a2)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import itertools as it
from operator import mul

allow_memoize_hash_failures = False


def safe_zip(*args):
  n = len(args[0])
  for arg in args[1:]:
    assert len(arg) == n, 'length mismatch: {}'.format(list(map(len, args)))
  return list(zip(*args))


def safe_map(f, *args):
  args = list(map(list, args))
  n = len(args[0])
  for arg in args[1:]:
    assert len(arg) == n, 'length mismatch: {}'.format(list(map(len, args)))
  return list(map(f, *args))


def unzip2(xys):
  xs = []
  ys = []
  for x, y in xys:
    xs.append(x)
    ys.append(y)
  return tuple(xs), tuple(ys)


def unzip3(xyzs):
  xs = []
  ys = []
  zs = []
  for x, y, z in xyzs:
    xs.append(x)
    ys.append(y)
    zs.append(z)
  return tuple(xs), tuple(ys), tuple(zs)


def concatenate(xs):
  return list(it.chain.from_iterable(xs))


def partial(fun, *args, **kwargs):
  wrapped = functools.partial(fun, *args, **kwargs)
  functools.update_wrapper(wrapped, fun)
  wrapped._bound_args = args
  return wrapped

class partialmethod(functools.partial):
  def __get__(self, instance, owner):
    if instance is None:
      return self
    else:
      return partial(self.func, instance,
                     *(self.args or ()), **(self.keywords or {}))

def curry(f):
  return partial(partial, f)

def toposort(end_node):
  child_counts = {}
  stack = [end_node]
  while stack:
    node = stack.pop()
    if id(node) in child_counts:
      child_counts[id(node)] += 1
    else:
      child_counts[id(node)] = 1
      stack.extend(node.parents)

  sorted_nodes = []
  childless_nodes = [end_node]
  while childless_nodes:
    node = childless_nodes.pop()
    sorted_nodes.append(node)
    for parent in node.parents:
      if child_counts[id(parent)] == 1:
        childless_nodes.append(parent)
      else:
        child_counts[id(parent)] -= 1

  return sorted_nodes[::-1]


def split_merge(predicate, xs):
  sides = list(map(predicate, xs))
  lhs = [x for x, s in zip(xs, sides) if s]
  rhs = [x for x, s in zip(xs, sides) if not s]
  def merge(new_lhs, new_rhs):
    out = []
    for s in sides:
      if s:
        out.append(new_lhs[0])
        new_lhs = new_lhs[1:]
      else:
        out.append(new_rhs[0])
        new_rhs = new_rhs[1:]
    assert not new_rhs
    assert not new_lhs
    return out

  return lhs, rhs, merge


def memoize(fun):
  cache = {}
  def memoized_fun(*args, **kwargs):
    key = (args, tuple(kwargs and sorted(kwargs.items())))
    try:
      return cache[key]
    except KeyError:
      ans = cache[key] = fun(*args, **kwargs)
      return ans
    except TypeError:
      if allow_memoize_hash_failures:
        return fun(*args, **kwargs)
      else:
        raise
  return memoized_fun


def prod(xs):
  return functools.reduce(mul, xs, 1)


class WrapHashably(object):
  def __init__(self, val):
    self.val = val

  def __hash__(self):
    return id(self.val)

  def __eq__(self, other):
    return self.val is other.val



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .util import curry, partial

def thunk(f):
  store = Store()
  def f_memoized():
    if not store:
      # TODO(dougalm): save/restore relevant environment state too
      store.store(f())
    return store.val

  return f_memoized

class StoreException(Exception): pass

class Store(object):
  def store(self, val):
    assert not self, "Store occupied"
    self._val = val

  @property
  def val(self):
    if not self:
      raise StoreException("Store empty")
    return self._val

  def __nonzero__(self):
    return hasattr(self, '_val')

  __bool__ = __nonzero__


@curry
def staged(f, *init_args):
  store = Store()
  def f_partial(*rest):
    ans, aux = f(*(init_args + rest))
    store.store(aux)
    return ans

  f_partial.__name__ = f.__name__ + "_staged"
  return f_partial, thunk(lambda: store.val)


class WrappedFun(object):
  def __init__(self, f, transforms, kwargs):
    self.f = f
    self.transforms = transforms
    self.kwargs = kwargs

  def wrap(self, *transformation):
    return WrappedFun(self.f, [transformation] + self.transforms, self.kwargs)

  def populate_stores(self, other):
    for (_, _, self_store), (_, _, other_store) in zip(self.transforms,
                                                       other.transforms):
      if self_store is not None:
        self_store.store(other_store.val)

  def call_wrapped(self, *args):
    stack = []
    for gen, gen_args, out_store in self.transforms:
      gen = gen(*(gen_args + tuple(args)))
      args = next(gen)
      stack.append((gen, out_store))

    del gen
    ans = self.f(*args, **self.kwargs)
    del args
    while stack:
      gen, out_store = stack.pop()
      ans = gen.send(ans)
      if out_store is not None:
        ans, side = ans
        out_store.store(side)

    return ans

  def __repr__(self):
    def transform_to_str(x):
      i, (gen, args, _) = x
      return "{}   : {}   {}".format(i, fun_name(gen), fun_name(args))
    transformation_stack = map(transform_to_str, enumerate(self.transforms))
    return "Wrapped function:\n" + '\n'.join(transformation_stack) + '\nCore: ' + fun_name(self.f) + '\n'

  def hashable_payload(self):
    return (self.f,
            tuple((gen, tuple(gen_args)) for gen, gen_args, _ in self.transforms),
            tuple(sorted(self.kwargs.items())))

  def __hash__(self):
    return hash(self.hashable_payload())

  def __eq__(self, other):
    return self.hashable_payload() == other.hashable_payload()

@curry
def transformation(gen, fun, *transformation_args):
  return fun.wrap(gen, transformation_args, None)

@curry
def transformation_with_aux(gen, fun, *transformation_args):
  out_store = Store()
  out_thunk = lambda: out_store.val
  return fun.wrap(gen, transformation_args, out_store), out_thunk

def fun_name(f):
  try:
    return f.__name__
  except:
    return str(f)

def wrap_init(f, kwargs={}):
  return WrappedFun(f, [], kwargs)


def memoize(call):
  cache = {}
  def memoized_fun(f, *args):
    key = (f, args)
    if key in cache:
      ans, f_prev = cache[key]
      f.populate_stores(f_prev)
    else:
      ans = call(f, *args)
      cache[key] = (ans, f)
    return ans

  return memoized_fun



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import namedtuple

import itertools as it

import numpy as onp

from six.moves import reduce

from .. import core
from ..core import Trace, Tracer, new_master, pack, AbstractTuple, JaxTuple
from ..abstract_arrays import ShapedArray, make_shaped_array, array_types
from ..ad_util import add_jaxvals_p
from ..linear_util import transformation, transformation_with_aux, wrap_init
from ..tree_util import register_pytree_node
from ..util import unzip2, partial, safe_map

map = safe_map


def batch(fun, in_vals, in_dims, out_dim_target):
  sizes = reduce(set.union, map(dimsize, in_dims, in_vals))
  if not sizes:
    return fun.call_wrapped(*in_vals), None  # no mapped dimensions
  elif len(sizes) == 1:
    out_val, out_dim = batch_transform(fun).call_wrapped(in_vals, in_dims)
    return moveaxis(sizes.pop(), out_dim_target, out_dim, out_val)
  else:
    raise TypeError("got inconsistent map dimension sizes: {}".format(sizes))


# TODO(mattjj,dougalm): could call batch_subtrace here (a bit redundant)
@transformation
def batch_transform(vals, dims):
  with new_master(BatchTrace) as master:
    trace = BatchTrace(master, core.cur_sublevel())
    in_tracers = map(partial(BatchTracer, trace), vals, dims)
    out_tracer = yield in_tracers
    out_tracer = trace.full_raise(out_tracer)
    out_val, out_dim = out_tracer.val, out_tracer.batch_dim
    del master
  yield (out_val, out_dim)


@transformation_with_aux
def batch_subtrace(master, dims, *vals):
  trace = BatchTrace(master, core.cur_sublevel())
  ans = yield map(partial(BatchTracer, trace), vals, dims)
  out_tracer = trace.full_raise(ans)
  out_val, out_dim = out_tracer.val, out_tracer.batch_dim
  yield out_val, out_dim


### tracer


class BatchTracer(Tracer):
  def __init__(self, trace, val, batch_dim):
    self.trace = trace
    self.val = val
    self.batch_dim = batch_dim

  @property
  def aval(self):
    batched_aval = get_aval(self.val)
    return remove_batch_dim_from_aval(self.batch_dim, batched_aval)

  def unpack(self):
    t = type(self.batch_dim)
    if t is tuple:
      batch_dims = self.batch_dim
    elif t is int:
      batch_dims = [self.batch_dim] * len(self.val)
    elif t is type(None):
      return tuple(self.val)
    else:
      raise TypeError(t)
    return map(partial(BatchTracer, self.trace), self.val, batch_dims)

  def full_lower(self):
    if self.batch_dim is None:
      return core.full_lower(self.val)
    else:
      return self

class BatchTrace(Trace):
  def pure(self, val):
    return BatchTracer(self, val, None)

  def lift(self, val):
    return BatchTracer(self, val, None)

  def sublift(self, val):
    return BatchTracer(self, val.val, val.batch_dim)

  def process_primitive(self, primitive, tracers, params):
    vals_in, dims_in = unzip2((t.val, t.batch_dim) for t in tracers)
    if all(bdim is None for bdim in dims_in):
      return primitive.bind(*vals_in, **params)
    else:
      batched_primitive = get_primitive_batcher(primitive)
      val_out, dim_out = batched_primitive(vals_in, dims_in, **params)
      return BatchTracer(self, val_out, dim_out)

  def process_call(self, call_primitive, f, tracers, params):
    vals, dims = unzip2((t.val, t.batch_dim) for t in tracers)
    if all(bdim is None for bdim in dims):
      return call_primitive.bind(f, *vals, **params)
    else:
      f, dim_out = batch_subtrace(f, self.master, dims)
      val_out = call_primitive.bind(f, *vals, **params)
      return BatchTracer(self, val_out, dim_out())

  def post_process_call(self, _, out_tracer):
    raise NotImplementedError  # TODO(mattjj,dougalm)

  def pack(self, tracers):
    vals = pack([t.val for t in tracers])
    batch_dim = tuple(t.batch_dim for t in tracers)
    return BatchTracer(self, vals, batch_dim)


### abstract values


def get_aval(x):
  if isinstance(x, Tracer):
    return raise_to_shaped(x.aval)
  else:
    return shaped_aval(x)

def shaped_aval(x):
  try:
    return pytype_aval_mappings[type(x)](x)
  except KeyError:
    raise TypeError("{} is not a valid type for batching".format(type(x)))

def raise_to_shaped(aval):
  if type(aval) is AbstractTuple:
    return AbstractTuple(map(raise_to_shaped, aval))
  elif isinstance(aval, ShapedArray):
    return ShapedArray(aval.shape, aval.dtype)
  else:
    raise TypeError(type(aval))

def remove_batch_dim_from_aval(bdim, aval):
  t = type(aval)
  if t is AbstractTuple:
    if type(bdim) is tuple:
      return AbstractTuple(map(remove_batch_dim_from_aval, bdim, aval))
    else:
      return AbstractTuple(map(partial(remove_batch_dim_from_aval, bdim), aval))
  elif t is ShapedArray:
    if bdim is None:
      return ShapedArray(aval.shape, aval.dtype)
    else:
      assert 0 <= bdim < aval.ndim
      unbatched_shape = tuple(onp.delete(aval.shape, bdim))
      return ShapedArray(unbatched_shape, aval.dtype)
  else:
    raise TypeError(t)

pytype_aval_mappings = {}

def shaped_jaxtuple(xs):
  return AbstractTuple(map(shaped_aval, xs))

pytype_aval_mappings[JaxTuple] = shaped_jaxtuple

for t in array_types:
  pytype_aval_mappings[t] = make_shaped_array


### primitives


primitive_batchers = {}

def get_primitive_batcher(p):
  try:
    return primitive_batchers[p]
  except KeyError:
    raise NotImplementedError(
        "Batching rule for '{}' not implemented".format(p))

def defvectorized(prim):
  primitive_batchers[prim] = partial(vectorized_batcher, prim)

def vectorized_batcher(prim, batched_args, batch_dims, **params):
  assert all(batch_dims[0] == bd for bd in batch_dims[1:])
  return prim.bind(*batched_args, **params), batch_dims[0]

def defbroadcasting(prim):
  primitive_batchers[prim] = partial(broadcast_batcher, prim)

def broadcast_batcher(prim, batched_args, batch_dims, **params):
  args = map(bdim_at_front, batched_args, batch_dims)
  ndim = max(map(onp.ndim, args))  # special case to handle scalar broadcasting
  args = map(partial(handle_scalar_broadcasting, ndim), args, batch_dims)
  return prim.bind(*args, **params), 0

def defreducer(prim):
  primitive_batchers[prim] = partial(reducer_batcher, prim)

def reducer_batcher(prim, batched_args, batch_dims, axes, **kwargs):
  operand, = batched_args
  bdim, = batch_dims
  axes = tuple(onp.where(onp.less(axes, bdim), axes, onp.add(axes, 1)))
  bdim_out = list(onp.delete(onp.arange(operand.ndim), axes)).index(bdim)
  if 'input_shape' in kwargs:
    kwargs['input_shape'] = operand.shape
  return prim.bind(operand, axes=axes, **kwargs), bdim_out

def add_batched(batched_args, batch_dims):
  xs, ys = batched_args
  bdx, bdy = batch_dims
  if bdx == bdy:
    return add_jaxvals_p.bind(xs, ys), bdx
  else:
    raise NotImplementedError  # TODO(mattjj)
primitive_batchers[add_jaxvals_p] = add_batched


### util


def bdim_at_front(x, bdim, broadcast_size=1):
  if bdim is None:
    return broadcast(x, broadcast_size) if onp.ndim(x) else x
  else:
    return move_dim_to_front(x, bdim)

def move_dim_to_front(x, dim):
  assert 0 <= dim < onp.ndim(x)
  if dim == 0:
    return x
  else:
    perm = (dim,) + tuple(range(dim)) + tuple(range(dim + 1, onp.ndim(x)))
    return x.transpose(perm)

def handle_scalar_broadcasting(nd, x, bdim):
  if bdim is None or nd == onp.ndim(x):
    return x
  else:
    return x.reshape(x.shape + (1,) * (nd - x.ndim))

def dimsize(dim, x):
  if type(x) is JaxTuple:
    return reduce(set.union, map(partial(dimsize, dim), x))
  elif type(dim) is int:
    return {x.shape[dim]}
  elif dim is None:
    return set()
  else:
    raise TypeError(type(dim))

def moveaxis(sz, dst, src, x):
  aval = get_aval(x)
  if type(aval) is AbstractTuple:
    if type(src) is tuple and type(dst) is tuple:
      return pack(map(partial(moveaxis, sz), dst, src, x))
    elif type(src) is tuple:
      return pack(map(partial(moveaxis, sz, dst), src, x))
    else:
      return pack(map(partial(moveaxis, sz, dst, src), x))
  elif isinstance(aval, ShapedArray):
    if src == dst:
      return x
    else:
      if src is None:
        x = broadcast(x, sz)
        src = 0
      if src == dst:
        return x
      else:
        perm = [i for i in range(onp.ndim(x)) if i != src]
        perm.insert(dst, src)
        return x.transpose(perm)
  else:
    raise TypeError(type(aval))

def broadcast(x, sz):
  try:
    return x.broadcast((sz,))
  except AttributeError:
    assert not isinstance(x, Tracer)
    return onp.broadcast_to(x, (sz,) + onp.shape(x))



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as onp
import six

from . import core
from . import ad_util
from . util import prod
from .lib import xla_bridge


def concretization_err_msg(fun):
  fname = getattr(fun, "__name__", fun)
  return ("Abstract value passed to function {} that requires a concrete value. "
          "Possibly tracing Python control flow using abstract values. "
          "If so, try using lax.cond or lax.while instead.").format(fname)

def concretization_function_error(fun):
  def error(self, *args):
    raise TypeError(concretization_err_msg(fun))
  return error


class UnshapedArray(core.AbstractValue):
  __slots__ = ['dtype']
  array_abstraction_level = 3

  def __init__(self, dtype):
    self.dtype = dtype

  def __eq__(self, other):
    return type(self) is type(other) and self.dtype == other.dtype

  def __hash__(self):
    return hash(str(self.dtype))

  def __repr__(self):
    return '{}({})'.format(self.__class__.__name__, self.str_short())

  _bool = _nonzero = concretization_function_error(bool)
  _float   = concretization_function_error(float)
  _int     = concretization_function_error(int)
  if six.PY2:
    _long    = concretization_function_error(long)
  _complex = concretization_function_error(complex)
  _hex     = concretization_function_error(hex)
  _oct     = concretization_function_error(oct)

  def at_least_vspace(self):
    return self

  def join(self, other):
    return self

  def str_short(self):
    return onp.dtype(self.dtype).name


class ShapedArray(UnshapedArray):
  __slots__ = ['shape']
  array_abstraction_level = 2

  def __init__(self, shape, dtype):
    self.dtype = onp.dtype(xla_bridge.canonicalize_dtype(dtype))
    self.shape = shape

  ndim = property(lambda self: len(self.shape))
  size = property(lambda self: prod(self.shape))

  def __eq__(self, other):
    return (type(self) is type(other)
            and self.dtype == other.dtype and self.shape == other.shape)

  def __hash__(self):
    return hash((self.shape, str(self.dtype)))

  def at_least_vspace(self):
    return self

  def join(self, other):
    if self.shape == other.shape and self.dtype == other.dtype:
      return self
    elif self.dtype == other.dtype:
      return UnshapedArray(self.dtype)
    else:
      raise TypeError(other)

  def str_short(self):
    dtypestr = onp.dtype(self.dtype).name
    shapestr = ','.join(map(str, self.shape))
    return '{}[{}]'.format(dtypestr, shapestr)

  def __len__(self):
    try:
      return self.shape[0]
    except IndexError:
      raise TypeError("len() of unsized object")  # same as numpy error


class ConcreteArray(ShapedArray):
  __slots__ = ['val']
  array_abstraction_level = 0

  def __init__(self, val):
    self.val = val
    self.shape = onp.shape(val)
    # canonicalized self.dtype doesn't necessarily match self.val
    self.dtype = onp.dtype(xla_bridge.canonicalize_dtype(onp.result_type(val)))
    assert self.dtype != onp.dtype('O')

  def __eq__(self, other):
    return (type(self) is type(other) and self.dtype == other.dtype
            and self.shape == other.shape and onp.all(self.val == other.val))

  def __hash__(self):
    return id(self.val)

  def at_least_vspace(self):
    return ShapedArray(self.shape, self.dtype)

  def join(self, other):
    if self == other:
      return self
    elif self.shape == other.shape and self.dtype == other.dtype:
      return ShapedArray(self.shape, self.dtype)
    elif self.dtype == other.dtype:
      return UnshapedArray(self.dtype)
    else:
      raise TypeError(other)

  def str_short(self):
    return str(self.val)


def make_shaped_array(x):
  dtype = xla_bridge.canonicalize_dtype(onp.result_type(x))
  return ShapedArray(onp.shape(x), dtype)

array_types = [onp.ndarray, onp.float64, onp.float32, onp.complex64,
               onp.int64, onp.int32, onp.bool_, onp.uint64, onp.uint32, float,
               int, bool]

for t in array_types:
  core.pytype_aval_mappings[t] = ConcreteArray



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools

import numpy as onp

from . import core
from . import linear_util as lu
from .core import pack, eval_jaxpr
from .api_util import flatten_fun, unflatten_fun, tree_to_jaxtuples
from .tree_util import (process_pytree, node_types, build_tree, PyTreeDef, leaf,
                        tree_map)
from .util import unzip2, unzip3, curry, partial, safe_map, WrapHashably
from .abstract_arrays import ShapedArray
from .interpreters import partial_eval as pe
from .interpreters import xla
from .interpreters import ad
from .interpreters import batching

map = safe_map

def jit(fun, static_argnums=()):
  def f_jitted(*args, **kwargs):
    f = lu.wrap_init(fun, kwargs)
    dyn_argnums = [i for i in range(len(args)) if i not in static_argnums]
    f, dyn_args = argnums_partial(f, dyn_argnums, args)
    args_flat, in_trees = unzip2(map(tree_to_jaxtuples, dyn_args))
    check_args(args_flat)
    flat_fun, out_tree = flatten_fun(f, in_trees)
    out_flat = xla.xla_call(flat_fun, *args_flat)
    return build_tree(out_tree(), out_flat)

  return f_jitted

def grad(fun, argnums=0):
  def grad_f(*args, **kwargs):
    f = lu.wrap_init(fun, kwargs)
    f_partial, dyn_args = argnums_partial(f, argnums, args)
    ans, vjp_py = vjp(f_partial, *dyn_args)
    check_scalar(ans)
    g = vjp_py(onp.ones((), onp.result_type(ans)))
    return g[0] if isinstance(argnums, int) else g


  return grad_f

@curry
def jacfwd(fun, x):
  fun = lu.wrap_init(fun)
  pushfwd = partial(jvp, fun, (x,))
  std_basis = onp.eye(onp.size(x)).reshape((-1,) + onp.shape(x)),
  y, jac_flat = vmap(pushfwd, std_basis, out_axes=(None, 0))
  return jac_flat.reshape(onp.shape(y) + onp.shape(x))

@curry
def jacrev(fun, x):
  fun = lu.wrap_init(fun)
  y, pullback = vjp(fun, x)
  std_basis = onp.eye(onp.size(y)).reshape((-1,) + onp.shape(y))
  jac_flat, = vmap(pullback, std_basis, out_axes=onp.ndim(y))
  return jac_flat.reshape(onp.shape(y) + onp.shape(x))

def hessian(fun):
  return jacfwd(jacrev(fun))

def vmap(fun, *args, **kwargs):
  in_axes = kwargs.pop("in_axes", 0)
  out_axes = kwargs.pop("out_axes", 0)
  if kwargs:
    msg = "vmap keyword args must be 'in_axes' and/or 'out_axes', got {}."
    raise TypeError(msg.format(', '.join(kwargs)))

  if type(in_axes) is int:
    in_axes = (in_axes,) * len(args)
  if not isinstance(fun, lu.WrappedFun):
    fun = lu.wrap_init(fun)
  in_flat, in_trees = unzip2(map(tree_to_jaxtuples, args))
  flat_fun, out_tree = flatten_fun(fun, in_trees)
  out_flat = batching.batch(flat_fun, in_flat, in_axes, out_axes)
  return build_tree(out_tree(), out_flat)

def jvp(fun, primals, tangents):
  def flatten_arg(primal, tangent):
    primal_jtuple, tree_def = tree_to_jaxtuples(primal)
    tangent_jtuple, tree_def_2 = tree_to_jaxtuples(tangent)
    assert tree_def == tree_def_2, (tree_def, tree_def_2)
    return primal_jtuple, tangent_jtuple, tree_def

  if not isinstance(fun, lu.WrappedFun):
    fun = lu.wrap_init(fun)
  ps_flat, ts_flat, in_trees = unzip3(map(flatten_arg, primals, tangents))
  flat_fun, out_tree = flatten_fun(fun, in_trees)
  out_primal, out_tangent = ad.jvp(flat_fun).call_wrapped(ps_flat, ts_flat)
  return (build_tree(out_tree(), out_primal), build_tree(out_tree(), out_tangent))

def linearize(traceable, *primals):
  fun = lu.wrap_init(traceable)
  primals_flat, in_trees = unzip2(map(tree_to_jaxtuples, primals))
  flat_fun, out_tree = flatten_fun(fun, in_trees)
  out_primal, out_pval, jaxpr, consts = ad.linearize(flat_fun, *primals_flat)
  out_tree = out_tree()
  out_primal_py = build_tree(out_tree, out_primal)
  lifted_jvp = partial(lift_linearized, jaxpr, consts, (in_trees, out_tree), out_pval)
  return out_primal_py, lifted_jvp

def lift_linearized(jaxpr, consts, io_tree, out_pval, py_args):
  def fun(*args):
    primals = pack(args) # doesn't matter what these are-they'll be ignored
    tangents = pack(args)
    _, ans = eval_jaxpr(jaxpr, consts, (), primals, tangents)
    return pe.merge_pvals(ans, out_pval)

  return unflatten_fun(fun, io_tree, *py_args)

def vjp(fun, *primals):
  if not isinstance(fun, lu.WrappedFun):
    fun = lu.wrap_init(fun)
  primals_flat, in_trees = unzip2(map(tree_to_jaxtuples, primals))
  check_args(primals_flat)
  flat_fun, out_tree = flatten_fun(fun, in_trees)
  out_primal, out_vjp = ad.vjp(flat_fun, primals_flat)
  out_tree = out_tree()
  out_primal_py = build_tree(out_tree, out_primal)
  ct_in_trees = [out_tree]
  ct_out_tree = PyTreeDef(node_types[tuple], None, in_trees)
  def out_vjp_packed(cotangent_in):
    return out_vjp(cotangent_in)
  vjp_py = partial(unflatten_fun, out_vjp_packed, (ct_in_trees, ct_out_tree))
  return out_primal_py, vjp_py


def trace_to_jaxpr(traceable, py_pvals, **kwargs):
  fun = lu.wrap_init(traceable)
  pvals, in_trees = unzip2(map(tree_to_pval_tuples, py_pvals))
  flat_fun, out_tree = flatten_fun(fun, in_trees)
  jaxpr, out_pval, consts = pe.trace_to_jaxpr(flat_fun, pvals, **kwargs)
  return jaxpr, consts, out_pval, (in_trees, out_tree())

def lift_jaxpr(jaxpr, consts, io_tree, pvals, py_args):
  def fun(*args):
    ans = eval_jaxpr(jaxpr, consts, (), *args)
    return pe.merge_pvals(ans, pvals)
  return unflatten_fun(fun, io_tree, *py_args)


device_put = jit(lambda x: x)
device_get_array = lambda x: x.copy() if type(x) is xla.DeviceArray else x
device_get = partial(tree_map, device_get_array)


@lu.transformation_with_aux
def flatten_fun(in_trees, *args, **kwargs):
  py_args = map(build_tree, in_trees, args)
  ans = yield py_args
  yield process_pytree(pack, ans)


def unflatten_fun(fun, io_tree, *py_args):
  in_trees_expected, out_tree = io_tree
  args, in_trees = unzip2(map(tree_to_jaxtuples, py_args))
  for i, (in_tree, expected) in enumerate(zip(in_trees, in_trees_expected)):
    if in_tree != expected:
      raise TypeError("Expected {}, got {}".format(expected, in_tree))

  ans = fun(*args)
  return build_tree(out_tree, ans)


tree_to_pval_tuples = partial(process_pytree, pe.pack_pvals)
tree_to_jaxtuples = partial(process_pytree, pack)


def argnums_partial(f, dyn_argnums, args):
  if isinstance(dyn_argnums, int):
    dyn_argnums = (dyn_argnums,)
  else:
    dyn_argnums = tuple(dyn_argnums)
  fixed_args = tuple([None if i in dyn_argnums else WrapHashably(arg)
                      for i, arg in enumerate(args)])
  dyn_args = [args[i] for i in dyn_argnums]
  return argnums_partial_(f, dyn_argnums, fixed_args), dyn_args

@lu.transformation
def argnums_partial_(dyn_argnums, fixed_args, *dyn_args):
  args = [None if arg is None else arg.val for arg in fixed_args]
  for i, arg in zip(dyn_argnums, dyn_args):
    args[i] = arg
  ans = yield args
  yield ans

def check_args(args):
  for arg in args:
    if not (isinstance(arg, core.Tracer) or core.valid_jaxtype(arg)):
      raise TypeError("Argument '{}' of type {} is not a valid JAX type"
                      .format(arg, type(arg)))

def check_scalar(x):
  msg = "Gradient only defined for scalar-output functions. Output was: {}".format
  try:
    aval = core.get_aval(x)
    if not (isinstance(aval, ShapedArray) and aval.shape == ()):
      raise TypeError(msg(x))
  except TypeError:
    raise TypeError(msg(x))



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.8
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.





from sys import version_info
if version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_pywrap_xla', [dirname(__file__)])
        except ImportError:
            import _pywrap_xla
            return _pywrap_xla
        if fp is not None:
            try:
                _mod = imp.load_module('_pywrap_xla', fp, pathname, description)
            finally:
                fp.close()
            return _mod
    _pywrap_xla = swig_import_helper()
    del swig_import_helper
else:
    import _pywrap_xla
del version_info
try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.


def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        if _newclass:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr_nondynamic(self, class_type, name, static=1):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    if (not static):
        return object.__getattr__(self, name)
    else:
        raise AttributeError(name)

def _swig_getattr(self, class_type, name):
    return _swig_getattr_nondynamic(self, class_type, name, 0)


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)

try:
    _object = object
    _newclass = 1
except AttributeError:
    class _object:
        pass
    _newclass = 0



def InitializeReplicaCount(replica_count):
    return _pywrap_xla.InitializeReplicaCount(replica_count)
InitializeReplicaCount = _pywrap_xla.InitializeReplicaCount

def InitializePlatformName(platform_name):
    return _pywrap_xla.InitializePlatformName(platform_name)
InitializePlatformName = _pywrap_xla.InitializePlatformName

def GetReplicaCount():
    return _pywrap_xla.GetReplicaCount()
GetReplicaCount = _pywrap_xla.GetReplicaCount

def TransferToInfeedLocal(literal):
    return _pywrap_xla.TransferToInfeedLocal(literal)
TransferToInfeedLocal = _pywrap_xla.TransferToInfeedLocal

def TransferToInfeedLocalReplica(literal, replica_number):
    return _pywrap_xla.TransferToInfeedLocalReplica(literal, replica_number)
TransferToInfeedLocalReplica = _pywrap_xla.TransferToInfeedLocalReplica

def TransferFromOutfeedLocalReplica(shape, replica_number):
    return _pywrap_xla.TransferFromOutfeedLocalReplica(shape, replica_number)
TransferFromOutfeedLocalReplica = _pywrap_xla.TransferFromOutfeedLocalReplica
class LocalShapedBuffer(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, LocalShapedBuffer, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, LocalShapedBuffer, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined")
    __repr__ = _swig_repr
    __swig_getmethods__["FromLiteral"] = lambda x: _pywrap_xla.LocalShapedBuffer_FromLiteral
    if _newclass:
        FromLiteral = staticmethod(_pywrap_xla.LocalShapedBuffer_FromLiteral)

    def ToLiteral(self):
        return _pywrap_xla.LocalShapedBuffer_ToLiteral(self)

    def shape(self):
        return _pywrap_xla.LocalShapedBuffer_shape(self)
    __swig_destroy__ = _pywrap_xla.delete_LocalShapedBuffer
    __del__ = lambda self: None
LocalShapedBuffer_swigregister = _pywrap_xla.LocalShapedBuffer_swigregister
LocalShapedBuffer_swigregister(LocalShapedBuffer)

def LocalShapedBuffer_FromLiteral(argument, shape_with_layout):
    return _pywrap_xla.LocalShapedBuffer_FromLiteral(argument, shape_with_layout)
LocalShapedBuffer_FromLiteral = _pywrap_xla.LocalShapedBuffer_FromLiteral

class LocalShapedBufferTuple(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, LocalShapedBufferTuple, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, LocalShapedBufferTuple, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined")
    __repr__ = _swig_repr

    def Release(self, i):
        return _pywrap_xla.LocalShapedBufferTuple_Release(self, i)

    def size(self):
        return _pywrap_xla.LocalShapedBufferTuple_size(self)
LocalShapedBufferTuple_swigregister = _pywrap_xla.LocalShapedBufferTuple_swigregister
LocalShapedBufferTuple_swigregister(LocalShapedBufferTuple)


def DestructureLocalShapedBufferTuple(local_shaped_buffer):
    return _pywrap_xla.DestructureLocalShapedBufferTuple(local_shaped_buffer)
DestructureLocalShapedBufferTuple = _pywrap_xla.DestructureLocalShapedBufferTuple
class XrtAllocation(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, XrtAllocation, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, XrtAllocation, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined")
    __repr__ = _swig_repr
    __swig_getmethods__["FromLiteral"] = lambda x: _pywrap_xla.XrtAllocation_FromLiteral
    if _newclass:
        FromLiteral = staticmethod(_pywrap_xla.XrtAllocation_FromLiteral)

    def ToLiteral(self):
        return _pywrap_xla.XrtAllocation_ToLiteral(self)

    def shape(self):
        return _pywrap_xla.XrtAllocation_shape(self)
XrtAllocation_swigregister = _pywrap_xla.XrtAllocation_swigregister
XrtAllocation_swigregister(XrtAllocation)

def XrtAllocation_FromLiteral(argument, session_target):
    return _pywrap_xla.XrtAllocation_FromLiteral(argument, session_target)
XrtAllocation_FromLiteral = _pywrap_xla.XrtAllocation_FromLiteral

class XrtAllocationTuple(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, XrtAllocationTuple, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, XrtAllocationTuple, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined")
    __repr__ = _swig_repr

    def Release(self, i):
        return _pywrap_xla.XrtAllocationTuple_Release(self, i)

    def size(self):
        return _pywrap_xla.XrtAllocationTuple_size(self)
XrtAllocationTuple_swigregister = _pywrap_xla.XrtAllocationTuple_swigregister
XrtAllocationTuple_swigregister(XrtAllocationTuple)


def DestructureXrtAllocationTuple(allocation, session_target):
    return _pywrap_xla.DestructureXrtAllocationTuple(allocation, session_target)
DestructureXrtAllocationTuple = _pywrap_xla.DestructureXrtAllocationTuple
class CompiledLocalComputation(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, CompiledLocalComputation, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, CompiledLocalComputation, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined")
    __repr__ = _swig_repr

    def Execute(self, argument_handles):
        return _pywrap_xla.CompiledLocalComputation_Execute(self, argument_handles)
    __swig_destroy__ = _pywrap_xla.delete_CompiledLocalComputation
    __del__ = lambda self: None
CompiledLocalComputation_swigregister = _pywrap_xla.CompiledLocalComputation_swigregister
CompiledLocalComputation_swigregister(CompiledLocalComputation)

class CompiledXrtComputation(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, CompiledXrtComputation, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, CompiledXrtComputation, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined")
    __repr__ = _swig_repr

    def Execute(self, argument_handles):
        return _pywrap_xla.CompiledXrtComputation_Execute(self, argument_handles)
CompiledXrtComputation_swigregister = _pywrap_xla.CompiledXrtComputation_swigregister
CompiledXrtComputation_swigregister(CompiledXrtComputation)

class LocalComputation(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, LocalComputation, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, LocalComputation, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined")
    __repr__ = _swig_repr

    def Compile(self, argument_shapes, build_options):
        return _pywrap_xla.LocalComputation_Compile(self, argument_shapes, build_options)

    def CompileForXrt(self, argument_shapes, session_target):
        return _pywrap_xla.LocalComputation_CompileForXrt(self, argument_shapes, session_target)

    def GetSerializedProto(self):
        return _pywrap_xla.LocalComputation_GetSerializedProto(self)

    def GetReturnValueShape(self):
        return _pywrap_xla.LocalComputation_GetReturnValueShape(self)
    __swig_destroy__ = _pywrap_xla.delete_LocalComputation
    __del__ = lambda self: None
LocalComputation_swigregister = _pywrap_xla.LocalComputation_swigregister
LocalComputation_swigregister(LocalComputation)

class LocalOp(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, LocalOp, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, LocalOp, name)

    def __init__(self, *args, **kwargs):
        raise AttributeError("No constructor defined")
    __repr__ = _swig_repr
    __swig_destroy__ = _pywrap_xla.delete_LocalOp
    __del__ = lambda self: None
LocalOp_swigregister = _pywrap_xla.LocalOp_swigregister
LocalOp_swigregister(LocalOp)

class LocalComputationBuilder(_object):
    __swig_setmethods__ = {}
    __setattr__ = lambda self, name, value: _swig_setattr(self, LocalComputationBuilder, name, value)
    __swig_getmethods__ = {}
    __getattr__ = lambda self, name: _swig_getattr(self, LocalComputationBuilder, name)
    __repr__ = _swig_repr

    def __init__(self, computation_name):
        this = _pywrap_xla.new_LocalComputationBuilder(computation_name)
        try:
            self.this.append(this)
        except Exception:
            self.this = this

    def SetOpMetadata(self, metadata):
        return _pywrap_xla.LocalComputationBuilder_SetOpMetadata(self, metadata)

    def ClearOpMetadata(self):
        return _pywrap_xla.LocalComputationBuilder_ClearOpMetadata(self)

    def Build(self):
        return _pywrap_xla.LocalComputationBuilder_Build(self)

    def BuildWithRoot(self, root):
        return _pywrap_xla.LocalComputationBuilder_BuildWithRoot(self, root)

    def Parameter(self, parameter_number, shape, name):
        return _pywrap_xla.LocalComputationBuilder_Parameter(self, parameter_number, shape, name)

    def GetShape(self, operand):
        return _pywrap_xla.LocalComputationBuilder_GetShape(self, operand)

    def GetReturnValueShape(self):
        return _pywrap_xla.LocalComputationBuilder_GetReturnValueShape(self)

    def Infeed(self, shape):
        return _pywrap_xla.LocalComputationBuilder_Infeed(self, shape)

    def Outfeed(self, operand, shape, outfeed_config):
        return _pywrap_xla.LocalComputationBuilder_Outfeed(self, operand, shape, outfeed_config)

    def ConstantLiteral(self, literal):
        return _pywrap_xla.LocalComputationBuilder_ConstantLiteral(self, literal)

    def Broadcast(self, operand, broadcast_sizes):
        return _pywrap_xla.LocalComputationBuilder_Broadcast(self, operand, broadcast_sizes)

    def BroadcastInDim(self, operand, shape, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_BroadcastInDim(self, operand, shape, broadcast_dimensions)

    def Pad(self, operand, padding_value, padding_config):
        return _pywrap_xla.LocalComputationBuilder_Pad(self, operand, padding_value, padding_config)

    def Reshape(self, operand, dimensions, new_sizes):
        return _pywrap_xla.LocalComputationBuilder_Reshape(self, operand, dimensions, new_sizes)

    def Collapse(self, operand, dimensions):
        return _pywrap_xla.LocalComputationBuilder_Collapse(self, operand, dimensions)

    def CrossReplicaSum(self, operand):
        return _pywrap_xla.LocalComputationBuilder_CrossReplicaSum(self, operand)

    def Slice(self, operand, start_indices, limit_indices, strides):
        return _pywrap_xla.LocalComputationBuilder_Slice(self, operand, start_indices, limit_indices, strides)

    def SliceInDim(self, operand, start_index, limit_index, stride, dimno):
        return _pywrap_xla.LocalComputationBuilder_SliceInDim(self, operand, start_index, limit_index, stride, dimno)

    def DynamicSlice(self, operand, start_indices, slice_sizes):
        return _pywrap_xla.LocalComputationBuilder_DynamicSlice(self, operand, start_indices, slice_sizes)

    def DynamicUpdateSlice(self, operand, update, start_indices):
        return _pywrap_xla.LocalComputationBuilder_DynamicUpdateSlice(self, operand, update, start_indices)

    def ConcatInDim(self, operands, dimension):
        return _pywrap_xla.LocalComputationBuilder_ConcatInDim(self, operands, dimension)

    def SelectAndScatterWithGeneralPadding(self, operand, select, window_dimensions, window_strides, padding, source, init_value, scatter):
        return _pywrap_xla.LocalComputationBuilder_SelectAndScatterWithGeneralPadding(self, operand, select, window_dimensions, window_strides, padding, source, init_value, scatter)

    def Tuple(self, elements):
        return _pywrap_xla.LocalComputationBuilder_Tuple(self, elements)

    def GetTupleElement(self, tuple_data, index):
        return _pywrap_xla.LocalComputationBuilder_GetTupleElement(self, tuple_data, index)

    def Dot(self, lhs, rhs):
        return _pywrap_xla.LocalComputationBuilder_Dot(self, lhs, rhs)

    def DotGeneral(self, lhs, rhs, dimension_numbers):
        return _pywrap_xla.LocalComputationBuilder_DotGeneral(self, lhs, rhs, dimension_numbers)

    def ConvGeneralDilated(self, lhs, rhs, window_strides, padding, lhs_dilation, rhs_dilation, dimension_numbers, feature_group_count):
        return _pywrap_xla.LocalComputationBuilder_ConvGeneralDilated(self, lhs, rhs, window_strides, padding, lhs_dilation, rhs_dilation, dimension_numbers, feature_group_count)

    def ConvertElementType(self, operand, new_element_type):
        return _pywrap_xla.LocalComputationBuilder_ConvertElementType(self, operand, new_element_type)

    def BitcastConvertType(self, operand, new_element_type):
        return _pywrap_xla.LocalComputationBuilder_BitcastConvertType(self, operand, new_element_type)

    def Call(self, local_computation, operands):
        return _pywrap_xla.LocalComputationBuilder_Call(self, local_computation, operands)

    def Transpose(self, operand, permutation):
        return _pywrap_xla.LocalComputationBuilder_Transpose(self, operand, permutation)

    def Rev(self, operand, dimensions):
        return _pywrap_xla.LocalComputationBuilder_Rev(self, operand, dimensions)

    def Map(self, operands, local_computation, dimensions):
        return _pywrap_xla.LocalComputationBuilder_Map(self, operands, local_computation, dimensions)

    def Reduce(self, operand, init_value, local_computation, dimensions_to_reduce):
        return _pywrap_xla.LocalComputationBuilder_Reduce(self, operand, init_value, local_computation, dimensions_to_reduce)

    def ReduceWindowWithGeneralPadding(self, operand, init_value, local_computation, window_dimensions, window_strides, base_dilations, window_dilations, padding):
        return _pywrap_xla.LocalComputationBuilder_ReduceWindowWithGeneralPadding(self, operand, init_value, local_computation, window_dimensions, window_strides, base_dilations, window_dilations, padding)

    def RngNormal(self, mu, sigma, shape):
        return _pywrap_xla.LocalComputationBuilder_RngNormal(self, mu, sigma, shape)

    def RngUniform(self, a, b, shape):
        return _pywrap_xla.LocalComputationBuilder_RngUniform(self, a, b, shape)

    def While(self, condition, body, init):
        return _pywrap_xla.LocalComputationBuilder_While(self, condition, body, init)

    def Conditional(self, predicate, true_operand, true_computation, false_operand, false_computation):
        return _pywrap_xla.LocalComputationBuilder_Conditional(self, predicate, true_operand, true_computation, false_operand, false_computation)

    def IsConstant(self, operand):
        return _pywrap_xla.LocalComputationBuilder_IsConstant(self, operand)

    def Sort(self, operand, dimension):
        return _pywrap_xla.LocalComputationBuilder_Sort(self, operand, dimension)

    def SortKeyVal(self, keys, values, dimension):
        return _pywrap_xla.LocalComputationBuilder_SortKeyVal(self, keys, values, dimension)

    def Select(self, lhs, rhs, ehs):
        return _pywrap_xla.LocalComputationBuilder_Select(self, lhs, rhs, ehs)

    def Clamp(self, lhs, rhs, ehs):
        return _pywrap_xla.LocalComputationBuilder_Clamp(self, lhs, rhs, ehs)

    def Eq(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Eq(self, lhs, rhs, broadcast_dimensions)

    def Ne(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Ne(self, lhs, rhs, broadcast_dimensions)

    def Ge(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Ge(self, lhs, rhs, broadcast_dimensions)

    def Gt(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Gt(self, lhs, rhs, broadcast_dimensions)

    def Lt(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Lt(self, lhs, rhs, broadcast_dimensions)

    def Le(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Le(self, lhs, rhs, broadcast_dimensions)

    def Add(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Add(self, lhs, rhs, broadcast_dimensions)

    def Sub(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Sub(self, lhs, rhs, broadcast_dimensions)

    def Mul(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Mul(self, lhs, rhs, broadcast_dimensions)

    def Div(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Div(self, lhs, rhs, broadcast_dimensions)

    def Rem(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Rem(self, lhs, rhs, broadcast_dimensions)

    def Max(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Max(self, lhs, rhs, broadcast_dimensions)

    def Min(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Min(self, lhs, rhs, broadcast_dimensions)

    def And(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_And(self, lhs, rhs, broadcast_dimensions)

    def Or(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Or(self, lhs, rhs, broadcast_dimensions)

    def Xor(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Xor(self, lhs, rhs, broadcast_dimensions)

    def ShiftLeft(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_ShiftLeft(self, lhs, rhs, broadcast_dimensions)

    def ShiftRightArithmetic(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_ShiftRightArithmetic(self, lhs, rhs, broadcast_dimensions)

    def ShiftRightLogical(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_ShiftRightLogical(self, lhs, rhs, broadcast_dimensions)

    def Atan2(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Atan2(self, lhs, rhs, broadcast_dimensions)

    def Pow(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Pow(self, lhs, rhs, broadcast_dimensions)

    def Complex(self, lhs, rhs, broadcast_dimensions):
        return _pywrap_xla.LocalComputationBuilder_Complex(self, lhs, rhs, broadcast_dimensions)

    def Not(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Not(self, operand)

    def Abs(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Abs(self, operand)

    def Exp(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Exp(self, operand)

    def Expm1(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Expm1(self, operand)

    def Floor(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Floor(self, operand)

    def Ceil(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Ceil(self, operand)

    def Round(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Round(self, operand)

    def Log(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Log(self, operand)

    def Log1p(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Log1p(self, operand)

    def Sign(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Sign(self, operand)

    def Cos(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Cos(self, operand)

    def Sin(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Sin(self, operand)

    def Tanh(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Tanh(self, operand)

    def IsFinite(self, operand):
        return _pywrap_xla.LocalComputationBuilder_IsFinite(self, operand)

    def Neg(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Neg(self, operand)

    def Sqrt(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Sqrt(self, operand)

    def Rsqrt(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Rsqrt(self, operand)

    def Square(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Square(self, operand)

    def Reciprocal(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Reciprocal(self, operand)

    def Erfc(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Erfc(self, operand)

    def Erf(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Erf(self, operand)

    def ErfInv(self, operand):
        return _pywrap_xla.LocalComputationBuilder_ErfInv(self, operand)

    def Lgamma(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Lgamma(self, operand)

    def Digamma(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Digamma(self, operand)

    def Acos(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Acos(self, operand)

    def Asin(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Asin(self, operand)

    def Atan(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Atan(self, operand)

    def Tan(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Tan(self, operand)

    def Acosh(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Acosh(self, operand)

    def Asinh(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Asinh(self, operand)

    def Atanh(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Atanh(self, operand)

    def Cosh(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Cosh(self, operand)

    def Sinh(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Sinh(self, operand)

    def Real(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Real(self, operand)

    def Imag(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Imag(self, operand)

    def Conj(self, operand):
        return _pywrap_xla.LocalComputationBuilder_Conj(self, operand)
    __swig_destroy__ = _pywrap_xla.delete_LocalComputationBuilder
    __del__ = lambda self: None
LocalComputationBuilder_swigregister = _pywrap_xla.LocalComputationBuilder_swigregister
LocalComputationBuilder_swigregister(LocalComputationBuilder)


def DeleteLocalShapedBuffer(local_shaped_buffer):
    return _pywrap_xla.DeleteLocalShapedBuffer(local_shaped_buffer)
DeleteLocalShapedBuffer = _pywrap_xla.DeleteLocalShapedBuffer

def DeleteXrtAllocation(allocation):
    return _pywrap_xla.DeleteXrtAllocation(allocation)
DeleteXrtAllocation = _pywrap_xla.DeleteXrtAllocation

def DeleteCompiledLocalComputation(computation):
    return _pywrap_xla.DeleteCompiledLocalComputation(computation)
DeleteCompiledLocalComputation = _pywrap_xla.DeleteCompiledLocalComputation

def DeleteCompiledXrtComputation(computation):
    return _pywrap_xla.DeleteCompiledXrtComputation(computation)
DeleteCompiledXrtComputation = _pywrap_xla.DeleteCompiledXrtComputation

def DeleteLocalComputation(computation):
    return _pywrap_xla.DeleteLocalComputation(computation)
DeleteLocalComputation = _pywrap_xla.DeleteLocalComputation
# This file is compatible with both classic and new-style classes.


